#############################################################################
# menubuttons.tcl Copyright 2004 r0lZ
# Menu VOB buttons infos.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

## For each VTS, the following variables are set:
## ::buttons::butinfo($vts,errorcode)	;# 0=OK
## ::buttons::butinfo($vts,errormsg)
## ::buttons::butinfo($vts,numvobids)
## For each VOB ID
## ::buttons::butinfo($vts,$vobid,numcellids)
## For each Cell ID
## ::buttons::butinfo($vts,$vobid,$cellid,navpack)	;# first NAV pack LBA for current VOB/Cell ID
## ::buttons::butinfo($vts,$vobid,$cellid,numbuts)	;# number of buttons (in first group)
## ::buttons::butinfo($vts,$vobid,$cellid,numgroups)
## ::buttons::butinfo($vts,$vobid,$cellid,startbutn)
## ::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn)
## ::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)
## For eacg group
## ::buttons::butinfo($vts,$vobid,$cellid,$group,grouptype)	;# 1=normal, 2=letterbox, 4=pan&scan (ored together)
## For each button
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,autoaction) ;# 0 or 1
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,coords)	;# Button position: {x_start y_start - x_end y_end}
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,udlr)		;# Adjacent buttons: {up down left right}
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,command)	;# Button command: {XX XX XX XX XX XX XX XX}
## and
## ::pgcs($vts,$pgc,cellpostable)		;# Cell Position Table, in binary

namespace eval ::menubuttons {


#############################################################################
## Procedure:  dump_pgc_menu_buttons
## args: if false, suppress the [hex codes] from the dump
proc dump_pgc_menu_buttons {vts numcells cellpostable indentstring args} {
	if {$numcells == 0} {return ""}
	if {$::tcl_platform(platform) == "macintosh"} {return ""}

	set txt "${indentstring}********** menu buttons commands:\n"
	for {set cell 0} {$cell < $numcells} {incr cell} {
		set idx [expr $cell * 4]
		set vobid [get_int cellpostable $idx 2]
		incr idx 3
		set cellid [get_int cellpostable $idx 1]
		if [info exists ::buttons::butinfo($vts,$vobid,$cellid,numbuts)] {
			set numbuts $::buttons::butinfo($vts,$vobid,$cellid,numbuts)
			if {$numbuts == 0} {continue}
			set txt "${txt}${indentstring}VOB ID $vobid,  Cell ID $cellid  (First NAV pack LBA=$::buttons::butinfo($vts,$vobid,$cellid,navpack))\n"

			set numgroups $::buttons::butinfo($vts,$vobid,$cellid,numgroups)
			set txt "${txt}${indentstring}$numgroups Group"
			if {$numgroups > 1} {
				set txt "${txt}s"
			}
			set txt "$txt,  Start button=$::buttons::butinfo($vts,$vobid,$cellid,startbutn),"
			set txt "$txt  Select button=$::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn),"
			set txt "$txt  Action button=$::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)\n"
			for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
				for {set group 1} {$group <= $numgroups} {incr group} {
					set cmdtxt "${indentstring}button [format "%2d" $butnum]: "
					set hex $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,command)
					if {$args == false} {
						set cmdtxt "${cmdtxt}[::disasm::disasm $hex] "
					} else {
						set cmdtxt "${cmdtxt}\[$hex\]  [::disasm::disasm $hex] "
					}
					if $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,autoaction) {
						set cmdtxt "$cmdtxt*** Autoaction "
					} else {
						set cmdtxt "$cmdtxt"
					}
					set cmdtext($group) $cmdtxt
				}
				set samecmds true
				if {$numgroups > 1} {
					if {$cmdtext(1) != $cmdtext(2)} {set samecmds false}
					if {$numgroups > 2} {
						if {$cmdtext(1) != $cmdtext(3) || $cmdtext(2) != $cmdtext(3)} {set samecmds false}
					}
				}
				if {! $samecmds} {
					for {set group 1} {$group <= $numgroups} {incr group} {
						set txt "${txt}$cmdtext($group)*** Group $group ([grouptype_descr $::buttons::butinfo($vts,$vobid,$cellid,$group,grouptype)])\n"
					}
				} else {
					set txt "${txt}$cmdtext(1)\n"
				}
			}
		}
	}
	return $txt
}

#############################################################################
## Procedure:  dump_all_button_infos
proc dump_all_button_infos {header} {
	set out "$header"
	set out "$out\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set out "$out\nProvider ID: \"[::utils::get_provider_ID]\""
	set out "$out\nNumber of VTS: $::pgcs(numvts)\n"

	# VIDEO_TS.IFO
	show_msg "Dumping VMGM buttons..."
	if {$::buttons::butinfo(0,errorcode) == 0} {
		set out "$out\nVMG (VIDEO_TS.VOB)"
		set out "$out\n------------------"
		set out "$out\n[dump_vts_button_infos 0 $::config(dump_hex_codes)]"
	}

	# VTS_xx_0.IFO
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		show_msg "Dumping VTS $vts buttons..."
		if {$::buttons::butinfo($vts,errorcode) == 0} {
			set out "$out\nVTS $vts (VTS_[format %02i $vts]_0.VOB)"
			set out "$out\n--------------------"
			set out "$out\n[dump_vts_button_infos $vts $::config(dump_hex_codes)]"
		}
	}
	show_msg ""
	return [string trimright $out "\n"]\n
}

#############################################################################
## Procedure:  dump_all_button_infos
## args: if false, suppress the [hex codes] from the dump
proc dump_vts_button_infos {vts args} {
	if {$vts == 0} {set head "VMGM"} else {set head "VTSM $vts"}
	set txt ""
	for {set vobid 1} {$vobid <= $::buttons::butinfo($vts,numvobids)} {incr vobid} {
		for {set cellid 1} {$cellid <= $::buttons::butinfo($vts,$vobid,numcellids)} {incr cellid} {
			set txt "$txt  $head,  VOB ID $vobid,  Cell ID $cellid  (First NAV pack LBA=$::buttons::butinfo($vts,$vobid,$cellid,navpack))\n"
			if {$::buttons::butinfo($vts,$vobid,$cellid,numbuts) == 0} {
				set txt "$txt  No buttons.\n"
			} else {
				set txt "$txt  $::buttons::butinfo($vts,$vobid,$cellid,numgroups) Group"
				if {$::buttons::butinfo($vts,$vobid,$cellid,numgroups) > 1} {
					set txt "${txt}s"
				}
				set txt "$txt,  Start but=$::buttons::butinfo($vts,$vobid,$cellid,startbutn),"
				set txt "$txt  Select but=$::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn),"
				set txt "$txt  Action but=$::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)\n"
				set numgroups $::buttons::butinfo($vts,$vobid,$cellid,numgroups)
				for {set butnum 1} {$butnum <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {incr butnum} {
					for {set group 1} {$group <= $numgroups} {incr group} {
						set cmdtxt "    button [format "%2d" $butnum]: "
						set hex $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,command)
						if {$args == false} {
							set cmdtxt "${cmdtxt}[::disasm::disasm $hex] "
						} else {
							set cmdtxt "${cmdtxt}\[$hex\]  [::disasm::disasm $hex] "
						}
						if $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,autoaction) {
							set cmdtxt "$cmdtxt*** Autoaction "
						} else {
							set cmdtxt "$cmdtxt"
						}
						set cmdtext($group) $cmdtxt
					}
					set samecmds true
					if {$numgroups > 1} {
						if {$cmdtext(1) != $cmdtext(2)} {set samecmds false}
						if {$numgroups > 2} {
							if {$cmdtext(1) != $cmdtext(3) || $cmdtext(2) != $cmdtext(3)} {set samecmds false}
						}
					}
					if {! $samecmds} {
						for {set group 1} {$group <= $numgroups} {incr group} {
							set txt "${txt}$cmdtext($group)*** Group $group ([grouptype_descr $::buttons::butinfo($vts,$vobid,$cellid,$group,grouptype)])\n"
						}
					} else {
						set txt "${txt}$cmdtext(1)\n"
					}
				}
			}
			set txt "$txt\n"
		}
	}
	set txt "$txt\n"
}

proc show_menu_buttons_info {} {
	if {! [info exists ::dvddir]} {bell; return}
	::main::showtext .buttoncmds "Menu buttons commands" [dump_all_button_infos "Menu buttons commands"]
}

proc grouptype_descr {type} {
	set descr ""
	if {[expr $type & 1] == 1} {set descr "${descr}+normal"}
	if {[expr $type & 2] == 2} {set descr "${descr}+letterbox"}
	if {[expr $type & 4] == 4} {set descr "${descr}+pan&scan"}
	return [string trim $descr "+"]
}

###################################################
## Display Buttons GUI
###################################################

proc display_current_menu_buttons {mode} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {$::scratchbook_active} {
		bell
		return 0
	}
	if {$::tcl_platform(platform) == "macintosh"} {
		show_msg "Menu buttons: Not available under Macintosh, except OS X"
		bell
		return 0
	}
	if {[lindex $::editor_cmds_from 0] != "menu"} {
		show_msg "Menu buttons: Available only for Menu PGCs"
		bell
		return 0
	}
	set indexlist [split [lindex $::editor_cmds_from 1] ","]

	set ::buttons::groupnum 1

	return [display_menu_buttons [lindex $indexlist 0] [lindex $indexlist 1] \
			[lindex $indexlist 2] 1024 $mode 1]
}

proc display_menu_buttons {vts lu pgc sprm8 mode show_cell} {
	set ::menubuttons::menubuttons_done 1
	set utiltitle "Menu buttons"
	set w .menubuttons

	set resize_divisor $::config(menubuttons_screen_divisor)

	if {! [info exists ::buttons::groupnum]} {set ::buttons::groupnum 1}
	if {$::menus($vts,$lu,$pgc,numcells) == 0} {
		if {$mode == "info"} {
			show_msg "Menu buttons: This menu has no video cells"
			bell
		}
		return 0
	}
	if {$::menus($vts,$lu,$pgc,cellpostable) == ""} {
		show_msg "Menu buttons: This menu has no video cell position table"
		bell
		return 0
	}
	if {$::buttons::butinfo($vts,errorcode) != 0} {
		show_msg "Menu buttons: Error when parsing Menu VOB file: $::buttons::butinfo($vts,errormsg)"
		bell
		return 0
	}

	set programmap $::menus($vts,$lu,$pgc,programmap)
	for {set p 1} {$p <= $::menus($vts,$lu,$pgc,numprograms)} {incr p} {
		set cell_program([get_int programmap [expr $p-1] 1]) $p
	}
	set old_pg 1
	for {set c 1} {$c <= $::menus($vts,$lu,$pgc,numcells)} {incr c} {
		if [info exists cell_program($c)] {
			set old_pg $cell_program($c)
		} else {
			set cell_program($c) $old_pg
		}
	}

	set buttonsfound false
	set cellpostable $::menus($vts,$lu,$pgc,cellpostable)

	set menulabels {}
	set menucommands {}
	set ::menubuttons::menutargets {}
	set cell 0
	for {set c 1} {$c <= $::menus($vts,$lu,$pgc,numcells)} {incr c} {
		set idx [expr ($c-1) * 4]
		set this_vobid [get_int cellpostable $idx 2]
		incr idx 3
		set this_cellid [get_int cellpostable $idx 1]
		if [info exists ::buttons::butinfo($vts,$this_vobid,$this_cellid,numbuts)] {
			set this_numbuts $::buttons::butinfo($vts,$this_vobid,$this_cellid,numbuts)
			lappend menulabels "Program $cell_program($c), Cell $c:  VOB ID $this_vobid, Cell ID $this_cellid ($this_numbuts buttons)"
			lappend menucommands "::menubuttons::display_cell $w $vts $lu $pgc $c $cell_program($c) $this_vobid $this_cellid 1024 $mode $resize_divisor"
			lappend ::menubuttons::menutargets "$cell_program($c) $c $this_vobid $this_cellid"
			if {$mode == "trace" && $cell == 0} {
				set cell $c
				set vobid $this_vobid
				set cellid $this_cellid
				set numbuts $this_numbuts
			}
			if {$this_numbuts != 0} {
				set buttonsfound true
 				if {$mode == "info" && $cell == 0} {
					set cell $c
					set vobid $this_vobid
					set cellid $this_cellid
					set numbuts $this_numbuts
				}
			}
		}
	}
	if {! $buttonsfound} {
		show_msg "Menu buttons: This menu has no buttons"
		if {$mode == "info"} bell
		return 0
	}
	if {$mode == "trace"} {
		set cell $show_cell
		set idx [expr ($cell-1) * 4]
		set vobid [get_int cellpostable $idx 2]
		incr idx 3
		set cellid [get_int cellpostable $idx 1]
		if [info exists ::buttons::butinfo($vts,$vobid,$cellid,numbuts)] {
			set numbuts $::buttons::butinfo($vts,$vobid,$cellid,numbuts)
			if {$numbuts == 0} {
				show_msg "Menu buttons: This menu has no buttons"
				if {$mode == "info"} bell
				return 0
			}
		}
	}

	### GUI
	button .b
	set font [.b cget -font]
	destroy .b
	set ::menubuttons::minxsize [font measure $font "00"]
	set ::menubuttons::minysize [expr [font metrics $font -ascent] - [font metrics $font -descent]+4]

	set nullimg {
		#define null_width 1
		#define null_height 1
		static unsigned char null_bits[] = {
			0x00};
	}
	image create bitmap ::menubuttons::bitmapnull -background {} -data $nullimg

	if [winfo exists $w] {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	if {$mode == "info"} {
		wm protocol $w WM_DELETE_WINDOW {set ::menubuttons::menubuttons_done 1}
	} else {
		wm protocol $w WM_DELETE_WINDOW bell
	}
	bind $w <Escape> {set ::menubuttons::menubuttons_done 1}
	if {$mode == "trace"} {
		bind $w <Return> {set ::menubuttons::menubuttons_done 1}
	}

	if {$mode == "trace"} {
		label $w.this -bd 1 -relief sunken -height 1 -textvariable ::menubuttons::lab1_var
	} else {
		menubutton $w.this -menu $w.this.select_cell -textvariable ::menubuttons::lab1_var -pady 1
		menu $w.this.select_cell
		for {set i 0} {$i < [llength $menulabels]} {incr i} {
			set breakat 30
			if {[llength $menulabels] > 90} {set breakat 40}
			if {[llength $menulabels] > 160} {set breakat 51}
			if {[expr $i % $breakat] == 0 && $i != 0} {set break 1} else {set break 0}
			$w.this.select_cell add radiobutton -label [lindex $menulabels $i] \
					-command [lindex $menucommands $i] -columnbreak $break \
					-variable ::menubuttons::cell_select_menu_item -value [expr $i+1]
		}
		set ::menubuttons::cell_select_menu_item $cell
	}
	pack  $w.this -fill x -padx 4 -pady 2 -ipady 2

	label $w.labinfo -bd 1 -relief sunken -height 1 \
		-textvariable ::menubuttons::lab2_var
	pack  $w.labinfo -fill x -padx 4 -pady 2 -ipady 2

	set scrndim [get_screen_dims menu $vts]
	pack [frame $w.video -bd 2 -relief ridge] -padx 4 -pady 4
	# used only to set window dimentions
	pack [canvas $w.video.f -width [expr round([lindex $scrndim 0] / $resize_divisor)] \
			-height [expr round([lindex $scrndim 1] / $resize_divisor)] -background gray -bd 0]

	frame $w.f2 -bd 2 -relief groove
	#if {$mode == "info"} {
		pack  $w.f2 -padx 4 -pady 2 -fill x -expand 1
	#}
	label $w.f2.labcmd -bd 1 -relief sunken -height 2 -textvariable ::menubuttons::command
	pack  $w.f2.labcmd -fill x -padx 4 -pady 2
	frame $w.f2.f
	pack  $w.f2.f -fill x -expand 1
	checkbutton $w.f2.f.auto -state disabled -text "Auto Action" -variable ::menubuttons::autoaction
	pack   $w.f2.f.auto -side left
	menubutton $w.f2.f.cp -text "Copy to Clipboard" -width 16 -menu $w.f2.f.cp.menu -pady 2
	pack   $w.f2.f.cp -side right -padx 3 -pady 1
	menu $w.f2.f.cp.menu
	$w.f2.f.cp.menu add command -label "Copy to clipboard" -command {::menubuttons::copy_current_command_to_clip 1}
	$w.f2.f.cp.menu add command -label "Append to clipboard" -command {::menubuttons::copy_current_command_to_clip 2}
	$w.f2.f.cp.menu add command -label "Copy as text" -command {::menubuttons::copy_current_command_to_clip 3}
	$w.f2.f.cp.menu add separator
	$w.f2.f.cp.menu add command -label "Append to scratchbook" -command {::menubuttons::append_current_command_to_scratchbook}
	if {$mode == "info"} {
		button $w.f2.f.follow -text "Jump to Target" -width 16 -pady 0	;# -command set later
		pack   $w.f2.f.follow -side right -pady 1
	}

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x -expand 1
	label  $w.fb.lab3 -text "Group"
	radiobutton $w.fb.g1 -variable ::buttons::groupnum -value 1 -text "1" -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell_program($cell) $vobid $cellid 1024 $mode $resize_divisor"
	radiobutton $w.fb.g2 -variable ::buttons::groupnum -value 2 -text "2" -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell_program($cell) $vobid $cellid 1024 $mode $resize_divisor"
	radiobutton $w.fb.g3 -variable ::buttons::groupnum -value 3 -text "3" -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell_program($cell) $vobid $cellid 1024 $mode $resize_divisor"
	if {$mode == "info"} {
		pack   $w.fb.lab3 -side left
		pack   $w.fb.g1 -side left
		pack   $w.fb.g2 -side left
		pack   $w.fb.g3 -side left
		set butlabtext "   Button"
	} else {
		set butlabtext "Button"
	}
	label  $w.fb.lab1 -text $butlabtext
	pack   $w.fb.lab1 -side left
	label  $w.fb.lab2 -textvariable ::buttons::default_button -padx 0 -width 3 -anchor w
	pack   $w.fb.lab2 -side left
	button $w.fb.prev -text "<" -width 2 -pady 0 \
		-command "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button - 1\] true $resize_divisor"
	pack   $w.fb.prev -side left
	button $w.fb.next -text ">" -width 2 -pady 0 \
		-command "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button + 1\] true $resize_divisor"
	pack   $w.fb.next -side left
	frame  $w.fb.pad1
	pack   $w.fb.pad1 -side left -fill x -expand 1
	if {$mode == "info"} {
		button $w.fb.ok -text "Close" -width 13 -command {set ::menubuttons::menubuttons_done 1}
	} else {
		button $w.fb.skip -text "Skip" -width 6 -command {set ::menubuttons::menubuttons_done 0}
		pack   $w.fb.skip -side left -padx 2
		button $w.fb.ok -text "Select" -width 6 -command {set ::menubuttons::menubuttons_done 1}
		::balloon::set_balloon $w.fb.ok "Use the currently\nselected button."
	}
	pack   $w.fb.ok -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [expr [winfo reqwidth $w] + 300] [winfo reqheight $w]
	if {! [info exists ::menubuttons::win_geometry]} {
		wm geometry $w $::config(menubuttons_win_geometry)
	} else {
		wm geometry $w $::menubuttons::win_geometry
	}

	display_cell $w $vts $lu $pgc $cell $cell_program($cell) $vobid $cellid $sprm8 $mode $resize_divisor

	wm deiconify $w
	if {$mode == "info"} {
		grab set $w	;# TODO: prohibit some functions when in Trace mode (ie. step)
	}

	set ::menubuttons::menubuttons_done 0
	tkwait variable ::menubuttons::menubuttons_done
	
	if {$::menubuttons::menubuttons_done == 1 && $::buttons::default_button > 0} {
		$w.video.f.b$::buttons::default_button configure -background green \
				-foreground black -relief sunken
		catch {image delete ::menubuttons::preview_but($::buttons::default_button)}
		if {$mode == "trace"} {
			after 500 "set ::menubuttons::menubuttons_done 1"
			tkwait variable ::menubuttons::menubuttons_done
		}
	}
	if {$::menubuttons::menubuttons_done == 0 && $::buttons::action_button > 0 \
			&& [winfo exists $w.video.f.b$::buttons::action_button]} {
		$w.video.f.b$::buttons::action_button configure -background green \
				-foreground black -relief sunken
		catch {image delete ::menubuttons::preview_but($::buttons::action_button)}
		if {$mode == "trace"} {
			after 500 "set ::menubuttons::menubuttons_done 0"
			tkwait variable ::menubuttons::menubuttons_done
		}
	}
	set ::menubuttons::win_geometry [wm geometry $w]
	destroy $w
	catch {image delete ::menubuttons::preview_bmp}
	for {set i 1} {$i <= 32} {incr i} {
		if [catch {image delete ::menubuttons::preview_but($i)}] break
	}
	if {$mode == "trace"} {
		set ::tr::sprm(8) [expr $::buttons::default_button * 1024]
	}
	if {$::menubuttons::menubuttons_done == 1} {
		return $::buttons::default_button
	} else {
		return $::buttons::action_button 
	}
}

proc display_cell {w vts lu pgc cell program vobid cellid sprm8 mode resize_divisor} {
	set numbuts $::buttons::butinfo($vts,$vobid,$cellid,numbuts)
	if {$vts == 0} {set vtslabel "VMGM"} else {set vtslabel "VTSM $vts"}
	set bcd [string range $::menus($vts,$lu,$pgc,cellpbtable) [expr (($cell - 1) * 24) + 4] [expr (($cell - 1) * 24) + 7]]
	set cell_time [::pgced::BCDtime $bcd false]
	if {[string range $cell_time 0 3] == "00:0"} {
		set cell_time [string range $cell_time 4 end]
	} elseif {[string range $cell_time 0 2] == "00:"} {
		set cell_time [string range $cell_time 3 end]
	} elseif {[string range $cell_time 0 0] == "0"} {
		set cell_time [string range $cell_time 1 end]
	}
	#set ::menubuttons::lab1_var "$vtslabel / LU $lu / PGC $pgc, PG $program, Cell $cell: VOB ID $vobid, Cell ID $cellid, $cell_time (NAV LBA=$::buttons::butinfo($vts,$vobid,$cellid,navpack))"
	set ::menubuttons::lab1_var "$vtslabel / LU $lu / PGC $pgc, PG $program, Cell $cell: VOB ID $vobid, Cell ID $cellid, $cell_time"
	::balloon::set_balloon $w.this "Cell's first NAV pack (or first NAV pack with buttons) LBA in VOB: $::buttons::butinfo($vts,$vobid,$cellid,navpack)"
	if {$numbuts == 0} {
		set ::menubuttons::lab2_var "No buttons"
		set scrndim [get_screen_dims menu $vts]
		destroy $w.video.f
		pack [canvas $w.video.f -width [expr round([lindex $scrndim 0] / $resize_divisor)] \
				-height [expr round([lindex $scrndim 1] / $resize_divisor)] -background gray -bd 0]
		display_preview $w.video.f $vts $lu $pgc $cell $vobid $cellid $resize_divisor
		set ::menubuttons::autoaction 1
		$w.f2.f.auto configure -disabledforeground [$w.fb.lab3 cget -disabledforeground]
		$w.fb.lab3 configure -state disabled
		$w.fb.g1 configure -state disabled
		$w.fb.g2 configure -state disabled
		$w.fb.g3 configure -state disabled
		$w.fb.prev configure -state disabled
		$w.fb.next configure -state disabled
		$w.fb.lab1 configure -state disabled
		$w.fb.lab2 configure -state disabled
		set ::buttons::default_button 0
		set ::menubuttons::command ""
		catch {$w.f2.labcmd configure -command bell -relief sunken}
		if {$mode == "info"} {
			set cellcmdnum [get_int ::menus($vts,$lu,$pgc,cellpbtable) [expr (($cell - 1) * 24) + 3] 1]
			if {$cellcmdnum > 0} {
				set cellcmds [get_type_cmds cell ::menus($vts,$lu,$pgc)]
				if {$cellcmdnum <= [llength $cellcmds]} {
					set command [string toupper [lindex $cellcmds [expr $cellcmdnum-1]]]
					set ::menubuttons::command "\[$command\]\n[::disasm::disasm $command]\n(Cell command)"
					$w.f2.f.follow configure -command \
						"::menubuttons::jump_to_target_inside_PGC $w $vts $lu $pgc $cell $program \$::menubuttons::command info -1"
					::main::highlight_jump_cmd menu $vts,$lu,$pgc $::menubuttons::command -1 false true
				}
			}
		}
		return
	}
	$w.f2.f.auto configure -disabledforeground black
	$w.fb.lab3 configure -state normal
	$w.fb.g1 configure -state normal
	$w.fb.prev configure -state normal
	$w.fb.next configure -state normal

	set ::buttons::action_button $::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)
	set ::menubuttons::lab2_var "Group $::buttons::groupnum ([grouptype_descr $::buttons::butinfo($vts,$vobid,$cellid,$::buttons::groupnum,grouptype)]), $numbuts buttons, force select button $::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn), force action button $::buttons::action_button"

	if {$::buttons::butinfo($vts,$vobid,$cellid,numgroups) < 3} {
		$w.fb.g3 configure -state disabled
		if {$::buttons::groupnum == 3} {set ::buttons::groupnum 2}
	} else {
		$w.fb.g3 configure -state normal -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell $program $vobid $cellid 1024 $mode $resize_divisor"
	}
	if {$::buttons::butinfo($vts,$vobid,$cellid,numgroups) < 2} {
		$w.fb.g2 configure -state disabled
		if {$::buttons::groupnum == 2} {set ::buttons::groupnum 1}
	} else {
		$w.fb.g2 configure -state normal -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell $program $vobid $cellid 1024 $mode $resize_divisor"
	}
	$w.fb.g1 configure -command "::menubuttons::display_cell $w $vts $lu $pgc $cell $program $vobid $cellid 1024 $mode $resize_divisor"
	$w.fb.prev configure -command \
		"::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button - 1\] true $resize_divisor"
	$w.fb.next configure -command \
		"::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button + 1\] true $resize_divisor"
	if {$mode == "info"} {
		$w.f2.f.follow configure -command \
			"::menubuttons::jump_to_target_inside_PGC $w $vts $lu $pgc $cell $program \$::menubuttons::command info -1"
	} else {
		set still_time [get_int ::menus($vts,$lu,$pgc,cellpbtable) [expr (($cell - 1) * 24) + 2] 1]
		if {$still_time < 255} {
			$w.fb.skip configure -state normal
			set txt "Wait to end of cell's playback\nwithout selecting any button.\n"
			if {$::buttons::action_button > 0} {
				set txt "${txt}(Will use action button $::buttons::action_button)"
			} else {
				set txt "${txt}(Will execute cell/post commands)"
			}
			::balloon::set_balloon $w.fb.skip $txt
			
		} else {
			$w.fb.skip configure -state disabled
			::balloon::set_balloon $w.fb.skip "This cell's Still Time is infinite.\nYou must select a button."
		}
	}

	set scrndim [get_screen_dims menu $vts]
	destroy $w.video.f
	pack [canvas $w.video.f -width [expr round([lindex $scrndim 0] / $resize_divisor)] \
			-height [expr round([lindex $scrndim 1] / $resize_divisor)] -background gray -bd 0]
	display_preview $w.video.f $vts $lu $pgc $cell $vobid $cellid $resize_divisor

	for {set but $numbuts} {$but > 0} {incr but -1} {
		display_button $w $vts $lu $pgc $vobid $cellid $but false $resize_divisor
	}
	if {$::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn) != 0} {
		if {$::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn) <= \
				$::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {
			set ::buttons::default_button $::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn)
		} else {
			set txt "Warning: The Force select button number is out of range!\n"
			set txt "Value (in the VOB file): $::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn)\n"
			set txt "Valid range for this menu: 1 to $::buttons::butinfo($vts,$vobid,$cellid,numbuts) (or 0 to use SPRM 8 instead)\n\n"
			set txt "${txt}The button 1 will be selected by default."
			tk_messageBox -title "PgcEdit: Show Menu Buttons" -icon warning \
					-message $txt -parent $w \
					-type ok
			set ::buttons::default_button 1
		}
	} elseif {$sprm8 >= 1024} {
		set button [expr $sprm8 / 1024]
		if {$button <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {
			set ::buttons::default_button $button
		} else {
			set txt "Warning: The SPRM 8: highlighted menu button number is out of range!\n"
			set txt "SPRM 8 button value: $sprm8 (button $button)\n"
			set txt "Valid button number range for this menu: 1 to $::buttons::butinfo($vts,$vobid,$cellid,numbuts)\n\n"
			set txt "${txt}The button 1 will be selected by default."
			tk_messageBox -title "PgcEdit: Show Menu Buttons" -icon warning \
					-message $txt -parent $w \
					-type ok
			set ::buttons::default_button 1
		}
	} else {
		set txt "Warning: The SPRM(8:highlighted menu button) value is 0!\n"
		set txt "Valid button number range for this menu: 1 to $::buttons::butinfo($vts,$vobid,$cellid,numbuts)\n\n"
		set txt "${txt}The button 1 will be selected by default."
		tk_messageBox -title "PgcEdit: Show Menu Buttons" -icon warning \
				-message $txt -parent $w \
				-type ok
		set ::buttons::default_button 1
	}
	display_button $w $vts $lu $pgc $vobid $cellid $::buttons::default_button true $resize_divisor
}

proc display_button {w vts lu pgc vobid cellid but default resize_divisor} {
	set group $::buttons::groupnum
	if {$but > 0 && $but <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {
		set coords $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,coords)
		set x_start [expr round([lindex $coords 0] / $resize_divisor)]
		set y_start [expr round([lindex $coords 1] / $resize_divisor)]
		set x_end   [expr round([lindex $coords 3] / $resize_divisor)]
		set y_end   [expr round([lindex $coords 4] / $resize_divisor)]
		set x_size  [expr $x_end - $x_start - 4]
		set y_size  [expr $y_end - $y_start - 4]
		update
		if  {$x_size < $::menubuttons::minxsize} {
			set x_size $::menubuttons::minxsize
			if {[expr $x_start + $::menubuttons::minxsize + 4] > [winfo reqwidth $w.video.f]} {
				set x_start [expr [winfo reqwidth $w.video.f] - ($x_size+4)]
			}
		}
		if  {$y_size < $::menubuttons::minysize} {
			set y_size $::menubuttons::minysize
			if {[expr $y_start + $::menubuttons::minysize + 4] > [winfo reqheight $w.video.f]} {
				set y_start [expr [winfo reqheight $w.video.f] - ($y_size+4)]
			}
		}

		if [winfo exists $w.video.f.b$but] {destroy $w.video.f.b$but}
		if {[catch {image inuse ::menubuttons::preview_bmp}] || \
					![image inuse ::menubuttons::preview_bmp]} {
			button $w.video.f.b$but -image ::menubuttons::bitmapnull -bd 1 -padx 0 -pady 0 \
					-compound center -highlightthickness 0 -relief raised -text $but \
					-width $x_size -height $y_size -command "::menubuttons::display_button \
						$w $vts $lu $pgc $vobid $cellid $but true $resize_divisor"
			if $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,autoaction) {
				$w.video.f.b$but configure -relief sunken
			}
		} else {
			image create photo ::menubuttons::preview_but($but) -width $x_size -height $y_size
			::menubuttons::preview_but($but) copy ::menubuttons::preview_bmp \
					-from [expr $x_start+1] [expr $y_start+1] \
					[expr $x_start+$x_size+1] [expr $y_start+$y_size+1]
			button $w.video.f.b$but -image ::menubuttons::preview_but($but) -bd 1 -padx 0 -pady 0 \
					-compound center -highlightthickness 0 -relief flat -text $but \
					-width $x_size -height $y_size -command "::menubuttons::display_button \
						$w $vts $lu $pgc $vobid $cellid $but true $resize_divisor"
		}
		::balloon::set_balloon $w.video.f.b$but \
				"Button $but\n[::disasm::disasm $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,command)]"
		if $default {
			set ::buttons::default_button $but
			if {[catch {image inuse ::menubuttons::preview_bmp}] || \
					![image inuse ::menubuttons::preview_bmp]} {
				for {set i 1} {$i <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {incr i} {
					$w.video.f.b$i configure -background [$w.video cget -background] -foreground red
				}
				$w.video.f.b$but configure -background lightgreen -foreground black
			} else {
				for {set i 1} {$i <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {incr i} {
					$w.video.f.b$i configure -background red -foreground red
				}
				$w.video.f.b$but configure -background green -foreground green
			}
			set ::menubuttons::autoaction $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,autoaction)
			set ::menubuttons::command "\[$::buttons::butinfo($vts,$vobid,$cellid,$group,$but,command)\]\n[::disasm::disasm $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,command)]"
			set udlr $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,udlr)
			bind $w <Up>    "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 0] true $resize_divisor"
			bind $w <Down>  "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 1] true $resize_divisor"
			bind $w <Left>  "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 2] true $resize_divisor"
			bind $w <Right> "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 3] true $resize_divisor"

			::main::highlight_jump_cmd menu $vts,$lu,$pgc $::menubuttons::command -1 false true
		}
		place $w.video.f.b$but -x $x_start -y $y_start
	}
}


proc display_preview {w vts lu pgc cell vobid cellid resize_divisor} {
	if {$::tcl_platform(platform)!="windows"} return
	if [catch {package require img::bmp 1.3}] return
	if {$vobid==0 || $cellid==0} return
	if {! [info exists ::buttons::butinfo($vts,$vobid,$cellid,navpack)]} return	;# TODO: Afficher quand mme une frame?

	set last_vobu_start [get_int ::menus($vts,$lu,$pgc,cellpbtable) [expr (($cell - 1) * 24) + 16] 4]

	set default_lba $::buttons::butinfo($vts,$vobid,$cellid,navpack)
#	# if the video is not a still frame, then show a frame somewhere later
#	# to avoid to show a black frame or a frame with no button text displayed.
#	# WARNING: In case of a still frame with sound, the displayed frame may be
#	# the first frame of the next pack!
#	if {$::buttons::butinfo($vts,$vobid,$cellid,numbuts)==0 && \
#			$last_vobu_start>$default_lba} {
#		#puts "$default_lba $last_vobu_start [expr $default_lba + (($last_vobu_start - $default_lba) / 10)]"
#		set default_lba [expr $default_lba + (($last_vobu_start - $default_lba) / 10)]
#	}

	if {$vts == 0} {
		set vobfile  [file nativename [file join $::dvddir "VIDEO_TS.VOB"]]
	} else {
		set vobfile  [file nativename [file join $::dvddir "VTS_[format %02d $vts]_0.VOB"]]
	}
	if {! [file exists $vobfile]} return
	set progname [file nativename [file join $::config(bindir) "dvd2bmp.exe"]]
	set tempfile [file nativename [file join $::config(tempdir) "pgcedit_preview.bmp"]]
	catch {file delete $tempfile}
	exec -- $progname "$vobfile" $default_lba $resize_divisor $resize_divisor "$tempfile"
	if {![file exists $tempfile]} return
	if {! [catch {image create photo ::menubuttons::preview_bmp -file $tempfile}]} {
		$w create image 0 0 -image ::menubuttons::preview_bmp -anchor nw
	}
	catch {file delete $tempfile}
}

proc get_screen_dims {type vtsnum} {
	if {$type == "menu"} {
		if {$vtsnum == 0} {
			set mat_table $::raw_sectors(0,VMGM_MAT)
			set vattr [get_int mat_table 0x100 2]
		} else {
			set mat_table $::raw_sectors($vtsnum,VTSI_MAT)
			set vattr [get_int mat_table 0x100 2]
		}
	} else {
		set mat_table $::raw_sectors($vtsnum,VTSI_MAT)
		set vattr [get_int mat_table 0x200 2]
	}
	if {[expr $vattr & 0X1000] == 0} {
		set h 480
	} else {
		set h 576
	}
	#switch [expr ($vattr & 0X000C) / 4] {
	#	0 {set w 720}
	#	1 {set w 704}
	#	2 {set w 352}
	#	3 {set w 352; set h [expr $h / 2]}
	#}
	if {[expr ($vattr & 0X000C) / 4] == 0} {
		set w 720
	} else {
		set w 704
	}
	return "$w $h"
}

proc copy_current_command_to_clip {mode} {
	if {$::menubuttons::command == ""} {
		bell
		return
	}
	switch $mode {
		1 {
			set cmd [list [string range $::menubuttons::command 1 23]]
		}
		2 {
			set old [clipboard get]
			set cmd [lappend old [string range $::menubuttons::command 1 23]]
		}
		3 {
			set cmd [string map {"\n" "  "} $::menubuttons::command]
		}
	}
	clipboard clear
	clipboard append $cmd
}

proc append_current_command_to_scratchbook {} {
	if {$::menubuttons::command == ""} {
		bell
		return
	}
	lappend ::scratchbook_buffer [string range $::menubuttons::command 1 23]
}

proc jump_to_target_inside_PGC {w vts lu pgc cell pg cmd mode linenum} {
	if {$::menubuttons::command == ""} {
		bell
		return 0
	}
	set cmd [string range $cmd 1 23]
	set jumpfound 0
	set targetpg   -1
	set targetcell -1
	# Link commands
	set cmdchk [lindex $cmd 0][string range [lindex $cmd 1] 1 1]
	switch $cmdchk {
		"201" {
			# Link*PG and Link*Cell
			set cmdchk2 [lindex $cmd 7]
			switch $cmdchk2 {
				"01" {
					# LinkTopCell
					set targetcell $cell
				}
				"02" {
					# LinkNextCell
					set targetcell [expr $cell + 1]
				}
				"03" {
					# LinkPrevCell
					set targetcell [expr $cell - 1]
				}
				"05" {
					# LinkTopPG
					set targetpg $pg
				}
				"06" {
					# LinkNextPG
					set targetpg [expr $pg + 1]
				}
				"07" {
					# LinkPrevPG
					set targetpg [expr $pg - 1]
				}
			}
		}
		"206" {
			# LinkPGN
			set targetpg [expr 0x[lindex $cmd 7]]
		}
		"207" {
			# LinkCN
			set targetcell [expr 0x[lindex $cmd 7]]
		}
	}
	# Set command with Link subcommand
	set cmdchk [string range [lindex $cmd 0] 0 0][string range [lindex $cmd 1] 1 1]
	switch $cmdchk {
		"61" - "71" {
			# Link*PG and Link*Cell
			set cmdchk2 [lindex $cmd 7]
			switch $cmdchk2 {
				"01" {
					# LinkTopCell
					set targetcell $cell
				}
				"02" {
					# LinkNextCell
					set targetcell [expr $cell + 1]
				}
				"03" {
					# LinkPrevCell
					set targetcell [expr $cell - 1]
				}
				"05" {
					# LinkTopPG
					set targetpg $pg
				}
				"06" {
					# LinkNextPG
					set targetpg [expr $pg + 1]
				}
				"07" {
					# LinkPrevPG
					set targetpg [expr $pg - 1]
				}
			}
		}
		"66" - "76" {
			# LinkPGN
			set targetpg [expr 0x[lindex $cmd 7]]
		}
		"67" - "77" {
			# LinkCN
			set targetcell [expr 0x[lindex $cmd 7]]
		}
	}
	# SetCLnk, CSetCLnk, CSetLnk commands
	set cmdchk [string range [lindex $cmd 0] 0 0]
	switch $cmdchk {
		"8" - "9" - "A" - "B" - "C" - "D" {
			# Link*PG and Link*Cell
			set cmdchk2 [lindex $cmd 7]
			switch $cmdchk2 {
				"01" {
					# LinkTopCell
					set targetcell $cell
				}
				"02" {
					# LinkNextCell
					set targetcell [expr $cell + 1]
				}
				"03" {
					# LinkPrevCell
					set targetcell [expr $cell - 1]
				}
				"05" {
					# LinkTopPG
					set targetpg $pg
				}
				"06" {
					# LinkNextPG
					set targetpg [expr $pg + 1]
				}
				"07" {
					# LinkPrevPG
					set targetpg [expr $pg - 1]
				}
			}
		}
	}
	if {$targetpg != -1} {
		set hl_btnn [expr (0x[lindex $cmd 6][lindex $cmd 7] / 1024) * 1024]
		for {set i 0} {$i < [llength $::menubuttons::menutargets]} {incr i} {
			set entry [lindex $::menubuttons::menutargets $i]
			if {$targetpg == [lindex $entry 0]} {
				set jumpfound 1
				set ::menubuttons::cell_select_menu_item [expr $i + 1]
				display_cell $w $vts $lu $pgc [lindex $entry 1] $targetpg [lindex $entry 2] [lindex $entry 3] $hl_btnn $mode
				break
			}
		}
		if {! $jumpfound} bell
	}
	if {$targetcell != -1} {
		set hl_btnn [expr (0x[lindex $cmd 6][lindex $cmd 7] / 1024) * 1024]
		for {set i 0} {$i < [llength $::menubuttons::menutargets]} {incr i} {
			set entry [lindex $::menubuttons::menutargets $i]
			if {$targetpg == [lindex $entry 1]} {
				set jumpfound 1
				set ::menubuttons::cell_select_menu_item [expr $i + 1]
				display_cell $w $vts $lu $pgc $targetcell [lindex $entry 0] [lindex $entry 2] [lindex $entry 3] $hl_btnn $mode
				break
			}
		}
		if {! $jumpfound} bell
	}

	if {! $jumpfound} {
		if {[::main::highlight_jump_cmd menu $vts,$lu,$pgc \[$cmd\] $linenum true true] == ""} {
			set jumpfound 0
		} else {
			set jumpfound 1
		}
		if $jumpfound {
			set ::menubuttons::menubuttons_done 1
		}
	}
	return $jumpfound
}

}
